from os import environ, getenv

from microcosm_logging.decorators import logger

from microcosm_sagemaker.decorators import metrics_observer, training_initializer
from microcosm_sagemaker.hyperparameters import get_hyperparameters


try:
    import wandb
except ImportError:
    pass


@training_initializer()
@metrics_observer()
@logger
class WeightsAndBiases:
    def __init__(self, graph):
        self.graph = graph
        self.testing = graph.metadata.testing
        self.project_name = graph.metadata.name.replace("_", "-")

    def _login(self):
        # TODO: Remove this line if devops come up with a solution
        # https://globality.atlassian.net/browse/DEVOPS-635
        if not getenv("WANDB_API_KEY"):
            environ["WANDB_API_KEY"] = self.graph.config.wandb.api_key

    def init(self):
        # Only initialize wandb if it is not a testing
        if not self.testing:
            self._login()
            wandb.init(
                project=self.project_name,
                config={
                    flattened_hyperparam: value
                    for flattened_hyperparam, value in get_hyperparameters(self.graph)
                }
            )
            self.logger.info("`weights & biases` was registered as a metric observer.")

    def log_timeseries(self, **kwargs):
        step = kwargs.pop("step")
        wandb.log(row=kwargs, step=step)
        return None

    def log_static(self, **kwargs):
        wandb.run.summary.update(kwargs)
        return None



from microcosm.config.model import Configuration
from microcosm.metadata import Metadata

from microcosm_sagemaker.constants import SagemakerPath


def load_default_microcosm_runserver_config(metadata: Metadata) -> Configuration:
    """
    Construct runserver default configuration.

    """

    config = Configuration(
        root_input_artifact_path=SagemakerPath.MODEL,
    )

    return config



import datetime
from typing import Optional

import boto3
from botocore.exceptions import ClientError, NoCredentialsError, NoRegionError
from microcosm_logging.decorators import logger

from microcosm_sagemaker.constants import CLOUDWATCH_MAX_DIMENSIONS
from microcosm_sagemaker.decorators import metrics_observer, training_initializer
from microcosm_sagemaker.hyperparameters import get_hyperparameters
from microcosm_sagemaker.metrics.sagemaker.models import LogMode, MetricUnit


@training_initializer()
@metrics_observer()
@logger
class SageMakerMetrics:
    def __init__(self, graph):
        self.graph = graph
        self.testing = graph.metadata.testing
        self.model_name = graph.metadata.name
        self.enabled = True

    @property
    def dimensions(self):
        return self._get_dimensions()

    def init(self):
        self.logger.info("`cloudwatch` was registered as a metric observer.")

    def _get_dimensions(self):
        # Metric dimensions allow us to analyze metric performance against the
        # hyperparameters of our model
        dimensions = [
            {
                "Name": flattened_hyperparam,
                "Value": str(value),
            }
            for flattened_hyperparam, value in get_hyperparameters(self.graph)
        ]

        if len(dimensions) > CLOUDWATCH_MAX_DIMENSIONS:
            self.logger.warning(
                f"The number of hyperparameters ({len(dimensions)}) is more than the maximum dimensions "
                f"allowed by `cloudwatch` ({CLOUDWATCH_MAX_DIMENSIONS}). "
                f"Truncating to first {CLOUDWATCH_MAX_DIMENSIONS} dimensions."
            )
            dimensions = dimensions[:CLOUDWATCH_MAX_DIMENSIONS]

        return dimensions

    def _metric_data(self, metric_name, metric_value, timestamp):
        metric_data = dict(
            MetricName=metric_name,
            Dimensions=self.dimensions,
            Value=metric_value,
            Unit=MetricUnit.NONE.name,
            StorageResolution=1,
        )

        if timestamp:
            metric_data.update(Timestamp=timestamp)

        return metric_data

    def _log_metric(self, log_mode: LogMode, **kwargs):

        timestamp: Optional[datetime.datetime]
        if log_mode == LogMode.TIMESERIES:
            timestamp = datetime.datetime.now()
        else:
            timestamp = None

        metric_data = [
            self._metric_data(metric_name, metric_value, timestamp)
            for metric_name, metric_value in kwargs.items()
        ]

        try:
            cloudwatch = boto3.client("cloudwatch")
            response = cloudwatch.put_metric_data(
                Namespace="/aws/sagemaker/" + self.model_name,
                MetricData=metric_data,
            )
        except (ClientError, NoCredentialsError, NoRegionError) as e:
            self.logger.warning("CloudWatch publishing disabled", extra=dict(metric_data=metric_data))  # type: ignore
            self.logger.warning(e)  # type: ignore
            # Disable cloudwatch logging if it fails to push metrics due to one of the above errors
            self.enabled = False
            response = None

        return response

    def log_static(self, **kwargs):
        if self.enabled:
            self._log_metric(LogMode.STATIC, **kwargs)

    def log_timeseries(self, **kwargs):
        if self.enabled:
            self._log_metric(LogMode.TIMESERIES, **kwargs)



"""
Main training CLI

"""
import json

from click import File, command, option
from microcosm.loaders import load_from_dict
from microcosm.object_graph import ObjectGraph

from microcosm_sagemaker.app_hooks import create_train_app
from microcosm_sagemaker.artifact import RootOutputArtifact
from microcosm_sagemaker.commands.options import input_data_option, output_artifact_option
from microcosm_sagemaker.exceptions import raise_sagemaker_exception
from microcosm_sagemaker.input_data import InputData


@command()
@option(
    "--configuration",
    type=File('r'),
    help="Manual import of configuration file, used for local testing",
)
@input_data_option()
@output_artifact_option()
@option(
    "--auto-evaluate/--no-auto-evaluate",
    default=True,
    help="Whether to automatically evaluate after the training has completed",
)
@option(
    "--testing/--no-testing",
    default=False,
    help="Whether the training is for testing purposes"
)
def main(configuration, input_data, output_artifact, auto_evaluate, testing):
    try:
        extra_config = json.load(configuration) if configuration else dict()

        graph = create_train_app(
            extra_loader=load_from_dict(extra_config),
            testing=testing,
        )

        run_train(
            graph=graph,
            input_data=input_data,
            root_output_artifact=output_artifact,
        )

        if auto_evaluate:
            run_auto_evaluate(graph, input_data)
    except Exception as e:
        raise_sagemaker_exception(e)


def run_train(
    graph: ObjectGraph,
    input_data: InputData,
    root_output_artifact: RootOutputArtifact,
) -> None:
    root_output_artifact.init()
    root_output_artifact.save_config(graph.config)

    graph.training_initializers.init()

    graph.bundle_and_dependencies_trainer(
        bundle=graph.active_bundle,
        input_data=input_data,
        root_output_artifact=root_output_artifact,
    )


def run_auto_evaluate(graph: ObjectGraph, input_data: InputData) -> None:
    graph.active_evaluation(graph.active_bundle, input_data)



from microcosm.config.model import Requirement
from microcosm.config.validation import zip_dicts
from microcosm.registry import get_defaults


class Hyperparameter(Requirement):
    """
    This class subclasses from Requirement and adds the `is_hyperparameter=True` flag.

    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.is_hyperparameter = True


def hyperparameter(default_value, parameter_type=None):
    """
    Fluent hyperparameter declaration.
    In most cases, the type is not needed and can be inferred from the default value.
    For example, to declare `epochs` as a hyperparameter, use:
    ```
    from microcosm.api import defaults
    from microcosm_sagemaker.hyperparameters import hyperparameter

    @defaults(
        epochs=hyperparameter(100)
    )
    class ClassifierBundle():
        ...
    ```

    """
    if not parameter_type:
        parameter_type = type(default_value)
    return Hyperparameter(type=parameter_type, default_value=default_value)


def get_hyperparameters(graph):
    """
    Given a graph, yields all of the hyperparameters in the config as
    `__`-separated keys.

    Consider the following example:

    ```
    @binding("ann_classifier_bundle")
    @defaults(
        no_of_epochs=hyperparameter(100),
        layer_sizes=dict(
            input_layer=10,
            hidden_layer=hyperparameter(20),
            output_layer=5,
        )
    )
    class ANNClassifierBundle():
        ...
    ```

    This functions yields the following:

    ```
    [
        "ann_classifier_bundle__no_of_epochs",
        "ann_classifier_bundle__layer_sizes__hidden_layer",
    ]
    ```

    """
    defaults = {
        key: get_defaults(graph.factory_for(key))
        for key, _ in graph.items()
    }

    for path, _, default, _, value in zip_dicts(defaults, graph.config):
        if isinstance(default, Hyperparameter):
            yield "__".join(path), value



from allennlp.predictors.predictor import Predictor


@Predictor.register("vanilla_predictor")
class VanillaPredictor(Predictor):
    """
    Predictor class that just does default allennlp prediction behavior.

    """
    pass
